/*
 * CLIF is a Load Injection Framework
 * Copyright (C) 2005,2006,2007 France Telecom
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option) any
 * later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Contact: clif@objectweb.org
 */

package org.objectweb.clif.isac.plugins;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.objectweb.clif.scenario.isac.plugin.SessionObjectAction;
import org.objectweb.clif.scenario.isac.plugin.TestAction;
import org.objectweb.clif.scenario.isac.exception.IsacRuntimeException;
import org.objectweb.clif.scenario.isac.plugin.ControlAction;
import org.objectweb.clif.scenario.isac.plugin.DataProvider;
import org.objectweb.clif.util.ClifClassLoader;

/**
 * Implementation of a session object for plugin ~CSVProvider~
 * Data provider taking external data from a CSV file.
 * Lines are read from a file as sequences of fields separated by a given separator character.
 * Each field may be referenced either by its index with an array notation [<i>n</i>]
 * (first index is 0) or by a name as optionally set at specimen session object creation.
 * Use <code>next</code> control to go to next line.
 * The sequence of line for each session object depends on two switches:
 * <ul>
 * <li>shared or not: when shared is enabled, the sequence of lines is shared by all
 * session objects associated to the same behavior (in other words copies of the same specimen).
 * When disabled, all session objects have their own sequence of lines. 
 * <li>loop or not: when loop is enabled, the sequence of lines automatically loops
 * to the first line when the last line is reached. When disabled, the <code>reset</code> control
 * must be used to loop; otherwise, an exception is thrown.
 * </ul>
 *  
 * @author Emmanuel Varoquaux
 * @author Bruno Dillenseger
 */
public class CSVProvider
	implements SessionObjectAction, TestAction, ControlAction, DataProvider
{
	static final int CONTROL_NEXT = 2;
	static final int CONTROL_RESET = 3;
	static final int TEST_ENDOFFILE = 0;
	static final int TEST_NOTENDOFFILE = 1;
	static final String PLUGIN_FIELDS = "fields";
	static final String PLUGIN_LOOP = "loop";
	static final String PLUGIN_SHARED = "shared";
	static final String PLUGIN_MACINTOSH_LINE_SEPARATOR = "macintosh_line_separator";
	static final String PLUGIN_SEPARATOR = "separator";
	static final String PLUGIN_FILENAME = "filename";
	static final String ENABLE_CBX = "enable;";
	static final String LINES_GET = "#";

	public static final char DEFAULT_SEPARATOR = ',';
	private static final Pattern pattern = Pattern.compile("\\[(\\d+)\\]");
	private List<String[]> values = new ArrayList<String[]>(); 
	private String[] fields;
	private int position = 0;
	private Boolean shared;
	private boolean loop;
	private CSVProvider parent;

	/**
	 * Constructor for specimen object.
	 *
	 * @param params key-value pairs for plugin parameters
	 */
	public CSVProvider(Hashtable params)
	{
		String tmp;
		boolean macintoshLineSeparator;
		char separator;

		if (params == null)
		{
			fields = null;
		}
		else
		{
			separator =
				(tmp = (String)params.get(PLUGIN_SEPARATOR)) != null && tmp.length() >= 1
				? tmp.charAt(0)
				: DEFAULT_SEPARATOR;
			fields =
				(tmp = (String)params.get(PLUGIN_FIELDS)) != null
				? split(new StringReader(tmp), separator, false)
				: null;
			macintoshLineSeparator =
				(tmp = (String)params.get(PLUGIN_MACINTOSH_LINE_SEPARATOR)) != null
				? tmp.equals(ENABLE_CBX)
				: false;
			shared = (tmp = (String)params.get(PLUGIN_SHARED)) != null
			? new Boolean(tmp.equals(ENABLE_CBX))
			: Boolean.FALSE;
			loop = (tmp = (String)params.get(PLUGIN_LOOP)) != null
			? tmp.equals(ENABLE_CBX)
			: false;
			try
			{
				String filename = (String)params.get(PLUGIN_FILENAME);
				Reader in = new InputStreamReader(ClifClassLoader.getClassLoader().getResourceAsStream(filename));
				String[] line;
				while ((line = split(in, separator, macintoshLineSeparator)) != null)
				{
					values.add(line);
				}
				position = -1;
			}
			catch (Exception e)
			{
				throw new Error(e);
			}
		}
	}

	/**
	 * Copy constructor (clone specimen object to get session object).
	 *
	 * @param so specimen object to clone
	 */
	private CSVProvider(CSVProvider so)
	{
		values = so.values;
		fields = so.fields;
		shared = so.shared;
		loop = so.loop;
		if (shared)
		{
			parent = so;
			doControl(CONTROL_NEXT, null);
		}
		else
		{
			position = 0;
		}
	}


	private String[] split(Reader enumeration, char separator, boolean macintoshLineSeparator)
	{
		final List<String> l = new LinkedList<String>();
		final StringBuilder s = new StringBuilder();
		int c;
		int state = 0;
		boolean again = true;

		/* CSV parser automaton */
		while (again)
		{
			try
			{
				c = enumeration.read();
			}
			catch (IOException ex)
			{
				c = -1;
			}
			if (c == -1)
			{
				l.add(new String(s));
				again = false;
			}
			else
			{
				switch (state)
				{
				case 0:
					switch (c)
					{
					case '\r':
						if (macintoshLineSeparator)
						{
							l.add(new String(s));
							again = false;
						}
						break;
					case '\n':
						if (!macintoshLineSeparator)
						{
							l.add(new String(s));
							again = false;
						}
						break;
					case '"':
						state = 2;
						break;
					default:
						if (c == separator)
						{
							l.add(new String(s));
							s.setLength(0);
						}
						else
						{
							s.append((char)c);
						}
					}
					break;
				// no more case 1
				case 2:
					switch (c)
					{
					case '"':
						state = 3;
						break;
					default:
						s.append((char)c);
					break;
					}
				case 3:
					switch (c) {
					case '"':
						s.append('"');
						state = 2;
						break;
					case '\r':
						state = 0;
						if (macintoshLineSeparator)
						{
							l.add(new String(s));
							again = false;
						}
						break;
					case '\n':
						state = 0;
						if (!macintoshLineSeparator)
						{
							l.add(new String(s));
							again = false;
						}
						break;
					default:
						state = 0;
						if (c == separator)
						{
							l.add(new String(s));
							s.setLength(0);
						}
						else
						{
							s.append((char)c);
						}
					}
				}
			}
		}
		return
			(l.size() == 1 && l.get(0).length() == 0)
			? null
			: l.toArray(new String[l.size()]);
	}


	////////////////////////////////////////
	// SessionObjectAction implementation //
	////////////////////////////////////////

	/**
	 * @see org.objectweb.clif.scenario.isac.plugin.SessionObjectAction#createNewSessionObject()
	 */
	public Object createNewSessionObject()
	{
		return new CSVProvider(this);
	}

	/**
	 * @see org.objectweb.clif.scenario.isac.plugin.SessionObjectAction#close()
	 */
	public void close()
	{
	}

	/**
	 * @see org.objectweb.clif.scenario.isac.plugin.SessionObjectAction#reset()
	 */
	public void reset()
	{
		if (shared)
		{
			doControl(CONTROL_NEXT, null);
		}
		else
		{
			position = 0;
		}
	}

	///////////////////////////////
	// TestAction implementation //
	///////////////////////////////

	/**
	 * @see org.objectweb.clif.scenario.isac.plugin.TestAction#doTest()
	 */
	public boolean doTest(int number, Map params)
	{
		switch (number)
		{
			case TEST_ENDOFFILE:
				return position >= values.size();
			case TEST_NOTENDOFFILE:
				return position < values.size();
			default:
				throw new Error("Unable to find this test in ~CSVProvider~ ISAC plugin: " + number);
		}
	}


	//////////////////////////////////
	// ControlAction implementation //
	//////////////////////////////////

	/**
	 * @see org.objectweb.clif.scenario.isac.plugin.ControlAction#doControl()
	 */
	public void doControl(int number, Map params)
	{
		switch (number)
		{
			case CONTROL_NEXT:
				if (shared)
				{
					synchronized(shared)
					{
						if (++parent.position == values.size() && loop)
						{
							parent.position = 0;
						}
						position = parent.position;
					}
				}
				else
				{
					if (++position == values.size() && loop)
					{
						position = 0;
					}
				}
				break;
			case CONTROL_RESET:
				if (shared)
				{
					synchronized(shared)
					{
						parent.position = 0;
						position = 0;
					}
				}
				else
				{
					position = 0;
				}
				break;
			default:
				throw new Error("Unable to find this control in ~CSVProvider~ ISAC plugin: " + number);
		}
	}


	/////////////////////////////////
	// DataProvider implementation //
	/////////////////////////////////

	/**
	 * @see org.objectweb.clif.scenario.isac.plugin.DataProvider#doGet()
	 */
	public String doGet(String var)
	{
		if (var.equals(LINES_GET))
		{
			return String.valueOf(values.size());
		}
		else
		{
			Matcher m;
			int i, n;

			if (position >= values.size())
			{
				throw new IsacRuntimeException("End of file reached in CSVProvider plug-in");
			}
			if (fields != null)
				for (i = 0, n = fields.length; i < n; i++)
					if (fields[i].equals(var))
						return (i < values.get(position).length) ? values.get(position)[i] : null;
			m = pattern.matcher(var);
			return
				m.matches()
				&& (i = Integer.parseInt(m.group(1))) < values.get(position).length
				? values.get(position)[i]
					: null;
		}
	}
}
